"""
automatic exposure time optimization
"""

import numpy as np

settings = {'name': 'automatic exposure', 'type': 'group', 'children': [
    {
        'name': 'target exposure', 'type': 'int', 'value': 80, 'limits': [1, 100], 'suffix': ' %',
        'tip': 'exposure target in % of full-scale',
    },
    {
        'name': 'allowed overexposed pixel', 'type': 'float', 'value': 0.1, 'limits': [0, 100], 'suffix': ' %',
        'tip': '% of pixels allowed to be overexposed (for instance defective pixel)',
    },
    {
        'name': 'number of iterations', 'type': 'int', 'value': 15, 'limits': [0, 100], 'suffix': ' frames',
        'tip': 'number of optimization iterations',
    },
    {
        'name': 'learning rate', 'type': 'float', 'value': 1.0, 'limits': [0.1, 1.9],
    },
    {
        'name': 'saturation rate', 'type': 'float', 'value': 8.0, 'limits': [1, 64],
    },
]}

class AutoExposure():
    def __init__(self, Settings):
        self.Settings = Settings
        self.init_optimize()

    def init_optimize(self):
        #print("DBG: init_optimize")
        self.iteration = 0

    def do_optimize(self, array, ExposureTime, n_bits, MinExposureTime, MaxExposureTime):
        """optimize exposure time

        @param array: image array
        @param ExposureTime: exposure time used to capture array
        @param n_bits: number of bits used in array
        @param MinExposureTime: minimum allowed exposure time
        @param MaxExposureTime: maximum allowed exposure time
        @return: (finished, new exposure time)
        """
        finished = True
        newExposureTime = ExposureTime
        if self.iteration < self.Settings.get('automatic exposure', 'number of iterations'):
            self.iteration += 1
            finished = False
            currentExposure = np.percentile(
                array,
                100 - self.Settings.get('automatic exposure', 'allowed overexposed pixel')
            )
            targetExposure = self.Settings.get('automatic exposure', 'target exposure') / 100 * (2 ** n_bits)
            #print(f'DBG: in do_optimize: {currentExposure=}, {targetExposure=}')
            if currentExposure < targetExposure:
                if ExposureTime <= 0:
                    newExposureTime = 0.01
                elif currentExposure <= 0:
                    # underexposed and saturated
                    newExposureTime = ExposureTime * self.Settings.get('automatic exposure', 'saturation rate')
                else:
                    newExposureTime = (ExposureTime +
                                       self.Settings.get('automatic exposure', 'learning rate') *
                                       (ExposureTime * targetExposure / currentExposure - ExposureTime)
                                       )
            else:
                n_saturated = (array >= ((2 ** n_bits) - 1)).sum()
                #print(f'DBG: in do_optimize: {n_saturated=}, {self.Settings.get("automatic exposure", "allowed overexposed pixel") / 100 * array.size}')
                if n_saturated > self.Settings.get('automatic exposure', 'allowed overexposed pixel') / 100 * array.size:
                    # overexposed and saturated
                    newExposureTime = ExposureTime / self.Settings.get('automatic exposure', 'saturation rate')
                else:
                    newExposureTime = (ExposureTime +
                                       self.Settings.get('automatic exposure', 'learning rate') *
                                       (ExposureTime * targetExposure / currentExposure - ExposureTime)
                                       )
        #print(f'DBG: in do_optimize: {ExposureTime=}, {newExposureTime=}')
        return finished, min(max(newExposureTime, MinExposureTime), MaxExposureTime)








"""
pifcap deticated image format
"""

import pickle

class Image:
    def __init__(self, array=None, metadata=None, format=None, comment=None):
        self.array = array
        self.metadata = metadata
        self.format =format
        self.comment = comment


    def get(self):
        Img = {
            "array": self.array,
            "format": self.format,
            "metadata": self.metadata,
            "comment": self.comment
        }
        return Img

    def estimate_FileSize(self):
        return len(pickle.dumps(self.get()))
    
    def save(self, filename):
        with open(filename, "wb") as fh:
            pickle.dump(self.get(), fh)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'src/pifcap/pifcap_ui.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1170, 746)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.widget_2 = QtWidgets.QWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(5)
        sizePolicy.setHeightForWidth(self.widget_2.sizePolicy().hasHeightForWidth())
        self.widget_2.setSizePolicy(sizePolicy)
        self.widget_2.setObjectName("widget_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.widget_2)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.splitter = QtWidgets.QSplitter(self.widget_2)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.widget_3 = QtWidgets.QWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget_3.sizePolicy().hasHeightForWidth())
        self.widget_3.setSizePolicy(sizePolicy)
        self.widget_3.setObjectName("widget_3")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.widget_3)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter_2 = QtWidgets.QSplitter(self.widget_3)
        self.splitter_2.setOrientation(QtCore.Qt.Vertical)
        self.splitter_2.setObjectName("splitter_2")
        self.ImageView_Preview = ImageView(self.splitter_2)
        self.ImageView_Preview.setObjectName("ImageView_Preview")
        self.plainTextEdit_Log = QtWidgets.QPlainTextEdit(self.splitter_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.plainTextEdit_Log.sizePolicy().hasHeightForWidth())
        self.plainTextEdit_Log.setSizePolicy(sizePolicy)
        self.plainTextEdit_Log.setObjectName("plainTextEdit_Log")
        self.verticalLayout_3.addWidget(self.splitter_2)
        self.widget = QtWidgets.QWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.scrollArea = QtWidgets.QScrollArea(self.widget)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents_3 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_3.setGeometry(QtCore.QRect(0, 0, 367, 691))
        self.scrollAreaWidgetContents_3.setObjectName("scrollAreaWidgetContents_3")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents_3)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.pushButton_Settings = QtWidgets.QPushButton(self.scrollAreaWidgetContents_3)
        self.pushButton_Settings.setObjectName("pushButton_Settings")
        self.verticalLayout_6.addWidget(self.pushButton_Settings)
        self.groupBox_3 = QtWidgets.QGroupBox(self.scrollAreaWidgetContents_3)
        self.groupBox_3.setObjectName("groupBox_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_3)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.comboBox_Camera = QtWidgets.QComboBox(self.groupBox_3)
        self.comboBox_Camera.setObjectName("comboBox_Camera")
        self.horizontalLayout.addWidget(self.comboBox_Camera)
        self.pushButton_ConnectDisconnect = QtWidgets.QPushButton(self.groupBox_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_ConnectDisconnect.sizePolicy().hasHeightForWidth())
        self.pushButton_ConnectDisconnect.setSizePolicy(sizePolicy)
        self.pushButton_ConnectDisconnect.setObjectName("pushButton_ConnectDisconnect")
        self.horizontalLayout.addWidget(self.pushButton_ConnectDisconnect)
        self.verticalLayout_6.addWidget(self.groupBox_3)
        self.groupBox_CameraSettings = QtWidgets.QGroupBox(self.scrollAreaWidgetContents_3)
        self.groupBox_CameraSettings.setObjectName("groupBox_CameraSettings")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_CameraSettings)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_6 = QtWidgets.QLabel(self.groupBox_CameraSettings)
        self.label_6.setObjectName("label_6")
        self.gridLayout_2.addWidget(self.label_6, 2, 0, 1, 1)
        self.doubleSpinBox_ExposureTime = QtWidgets.QDoubleSpinBox(self.groupBox_CameraSettings)
        self.doubleSpinBox_ExposureTime.setDecimals(4)
        self.doubleSpinBox_ExposureTime.setMaximum(10000.0)
        self.doubleSpinBox_ExposureTime.setSingleStep(0.01)
        self.doubleSpinBox_ExposureTime.setStepType(QtWidgets.QAbstractSpinBox.AdaptiveDecimalStepType)
        self.doubleSpinBox_ExposureTime.setObjectName("doubleSpinBox_ExposureTime")
        self.gridLayout_2.addWidget(self.doubleSpinBox_ExposureTime, 2, 1, 1, 1)
        self.doubleSpinBox_Gain = QtWidgets.QDoubleSpinBox(self.groupBox_CameraSettings)
        self.doubleSpinBox_Gain.setObjectName("doubleSpinBox_Gain")
        self.gridLayout_2.addWidget(self.doubleSpinBox_Gain, 1, 1, 1, 1)
        self.comboBox_RawMode = QtWidgets.QComboBox(self.groupBox_CameraSettings)
        self.comboBox_RawMode.setObjectName("comboBox_RawMode")
        self.gridLayout_2.addWidget(self.comboBox_RawMode, 0, 0, 1, 4)
        self.label_5 = QtWidgets.QLabel(self.groupBox_CameraSettings)
        self.label_5.setObjectName("label_5")
        self.gridLayout_2.addWidget(self.label_5, 1, 0, 1, 1)
        self.checkBox_OptimizeExposure = QtWidgets.QCheckBox(self.groupBox_CameraSettings)
        self.checkBox_OptimizeExposure.setObjectName("checkBox_OptimizeExposure")
        self.gridLayout_2.addWidget(self.checkBox_OptimizeExposure, 2, 2, 1, 1)
        self.verticalLayout_6.addWidget(self.groupBox_CameraSettings)
        self.groupBox_4 = QtWidgets.QGroupBox(self.scrollAreaWidgetContents_3)
        self.groupBox_4.setObjectName("groupBox_4")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox_4)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.checkBox_PreviewFlipH = QtWidgets.QCheckBox(self.groupBox_4)
        self.checkBox_PreviewFlipH.setObjectName("checkBox_PreviewFlipH")
        self.gridLayout_3.addWidget(self.checkBox_PreviewFlipH, 0, 1, 1, 1)
        self.checkBox_PreviewFlipV = QtWidgets.QCheckBox(self.groupBox_4)
        self.checkBox_PreviewFlipV.setObjectName("checkBox_PreviewFlipV")
        self.gridLayout_3.addWidget(self.checkBox_PreviewFlipV, 0, 2, 1, 1)
        self.comboBox_PreviewRot = QtWidgets.QComboBox(self.groupBox_4)
        self.comboBox_PreviewRot.setObjectName("comboBox_PreviewRot")
        self.comboBox_PreviewRot.addItem("")
        self.comboBox_PreviewRot.addItem("")
        self.comboBox_PreviewRot.addItem("")
        self.comboBox_PreviewRot.addItem("")
        self.gridLayout_3.addWidget(self.comboBox_PreviewRot, 0, 0, 1, 1)
        self.comboBox_RawPreviewMode = QtWidgets.QComboBox(self.groupBox_4)
        self.comboBox_RawPreviewMode.setEnabled(True)
        self.comboBox_RawPreviewMode.setObjectName("comboBox_RawPreviewMode")
        self.comboBox_RawPreviewMode.addItem("")
        self.comboBox_RawPreviewMode.addItem("")
        self.comboBox_RawPreviewMode.addItem("")
        self.comboBox_RawPreviewMode.addItem("")
        self.comboBox_RawPreviewMode.addItem("")
        self.gridLayout_3.addWidget(self.comboBox_RawPreviewMode, 1, 0, 1, 1)
        self.checkBox_Saturation = QtWidgets.QCheckBox(self.groupBox_4)
        self.checkBox_Saturation.setEnabled(True)
        self.checkBox_Saturation.setChecked(True)
        self.checkBox_Saturation.setObjectName("checkBox_Saturation")
        self.gridLayout_3.addWidget(self.checkBox_Saturation, 1, 1, 1, 2)
        self.verticalLayout_6.addWidget(self.groupBox_4)
        self.groupBox = QtWidgets.QGroupBox(self.scrollAreaWidgetContents_3)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout.setObjectName("gridLayout")
        self.label_3 = QtWidgets.QLabel(self.groupBox)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 0, 2, 1, 1)
        self.label_RecordedImageCounter = QtWidgets.QLabel(self.groupBox)
        self.label_RecordedImageCounter.setObjectName("label_RecordedImageCounter")
        self.gridLayout.addWidget(self.label_RecordedImageCounter, 5, 2, 1, 3)
        self.label_2 = QtWidgets.QLabel(self.groupBox)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 3, 0, 1, 1)
        self.pushButton_StopRec = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_StopRec.setEnabled(False)
        self.pushButton_StopRec.setObjectName("pushButton_StopRec")
        self.gridLayout.addWidget(self.pushButton_StopRec, 5, 5, 1, 2)
        self.pushButton_Folder = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_Folder.setObjectName("pushButton_Folder")
        self.gridLayout.addWidget(self.pushButton_Folder, 0, 0, 1, 2)
        self.doubleSpinBox_TimeLapse = QtWidgets.QDoubleSpinBox(self.groupBox)
        self.doubleSpinBox_TimeLapse.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.doubleSpinBox_TimeLapse.sizePolicy().hasHeightForWidth())
        self.doubleSpinBox_TimeLapse.setSizePolicy(sizePolicy)
        self.doubleSpinBox_TimeLapse.setMaximumSize(QtCore.QSize(70, 16777215))
        self.doubleSpinBox_TimeLapse.setPrefix("")
        self.doubleSpinBox_TimeLapse.setDecimals(1)
        self.doubleSpinBox_TimeLapse.setMaximum(1000000.0)
        self.doubleSpinBox_TimeLapse.setObjectName("doubleSpinBox_TimeLapse")
        self.gridLayout.addWidget(self.doubleSpinBox_TimeLapse, 3, 5, 1, 1)
        self.label_RecordingInfos = QtWidgets.QLabel(self.groupBox)
        self.label_RecordingInfos.setText("")
        self.label_RecordingInfos.setObjectName("label_RecordingInfos")
        self.gridLayout.addWidget(self.label_RecordingInfos, 4, 0, 1, 7)
        self.pushButton_StartRec = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_StartRec.setObjectName("pushButton_StartRec")
        self.gridLayout.addWidget(self.pushButton_StartRec, 5, 0, 1, 2)
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setEnabled(False)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 3, 4, 1, 1)
        self.lineEdit_FileNamePrefix = QtWidgets.QLineEdit(self.groupBox)
        self.lineEdit_FileNamePrefix.setObjectName("lineEdit_FileNamePrefix")
        self.gridLayout.addWidget(self.lineEdit_FileNamePrefix, 0, 3, 1, 4)
        self.lineEdit_ImageComment = QtWidgets.QLineEdit(self.groupBox)
        self.lineEdit_ImageComment.setObjectName("lineEdit_ImageComment")
        self.gridLayout.addWidget(self.lineEdit_ImageComment, 2, 1, 1, 6)
        self.spinBox_nImagesToRecord = QtWidgets.QSpinBox(self.groupBox)
        self.spinBox_nImagesToRecord.setMaximumSize(QtCore.QSize(70, 16777215))
        self.spinBox_nImagesToRecord.setMinimum(1)
        self.spinBox_nImagesToRecord.setMaximum(999999)
        self.spinBox_nImagesToRecord.setObjectName("spinBox_nImagesToRecord")
        self.gridLayout.addWidget(self.spinBox_nImagesToRecord, 3, 1, 1, 3)
        self.label_7 = QtWidgets.QLabel(self.groupBox)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 1, 0, 1, 1)
        self.comboBox_FrameType = QtWidgets.QComboBox(self.groupBox)
        self.comboBox_FrameType.setObjectName("comboBox_FrameType")
        self.comboBox_FrameType.addItem("")
        self.comboBox_FrameType.addItem("")
        self.comboBox_FrameType.addItem("")
        self.comboBox_FrameType.addItem("")
        self.gridLayout.addWidget(self.comboBox_FrameType, 1, 1, 1, 6)
        self.verticalLayout_6.addWidget(self.groupBox)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_6.addItem(spacerItem)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents_3)
        self.verticalLayout_5.addWidget(self.scrollArea)
        self.verticalLayout_4.addWidget(self.splitter)
        self.verticalLayout_2.addWidget(self.widget_2)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1170, 28))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.action_DarkNoiseMeasurement = QtWidgets.QAction(MainWindow)
        self.action_DarkNoiseMeasurement.setObjectName("action_DarkNoiseMeasurement")

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.pushButton_Settings.setText(_translate("MainWindow", "program settings"))
        self.groupBox_3.setTitle(_translate("MainWindow", "Camera"))
        self.pushButton_ConnectDisconnect.setText(_translate("MainWindow", "connect"))
        self.groupBox_CameraSettings.setTitle(_translate("MainWindow", "Camera Settings"))
        self.label_6.setText(_translate("MainWindow", "exposure time:"))
        self.doubleSpinBox_ExposureTime.setSuffix(_translate("MainWindow", " s"))
        self.label_5.setText(_translate("MainWindow", "gain:"))
        self.checkBox_OptimizeExposure.setText(_translate("MainWindow", "optimize"))
        self.groupBox_4.setTitle(_translate("MainWindow", "Display"))
        self.checkBox_PreviewFlipH.setText(_translate("MainWindow", "flip H"))
        self.checkBox_PreviewFlipV.setText(_translate("MainWindow", "flip V"))
        self.comboBox_PreviewRot.setItemText(0, _translate("MainWindow", "0deg"))
        self.comboBox_PreviewRot.setItemText(1, _translate("MainWindow", "90deg"))
        self.comboBox_PreviewRot.setItemText(2, _translate("MainWindow", "180deg"))
        self.comboBox_PreviewRot.setItemText(3, _translate("MainWindow", "270deg"))
        self.comboBox_RawPreviewMode.setItemText(0, _translate("MainWindow", "luminance"))
        self.comboBox_RawPreviewMode.setItemText(1, _translate("MainWindow", "red"))
        self.comboBox_RawPreviewMode.setItemText(2, _translate("MainWindow", "green 1"))
        self.comboBox_RawPreviewMode.setItemText(3, _translate("MainWindow", "green 2"))
        self.comboBox_RawPreviewMode.setItemText(4, _translate("MainWindow", "blue"))
        self.checkBox_Saturation.setText(_translate("MainWindow", "saturation"))
        self.groupBox.setTitle(_translate("MainWindow", "Recording"))
        self.label_3.setText(_translate("MainWindow", "Prefix:"))
        self.label_RecordedImageCounter.setText(_translate("MainWindow", "0 images saved"))
        self.label_2.setText(_translate("MainWindow", "Comment:"))
        self.label_4.setText(_translate("MainWindow", "# Images:"))
        self.pushButton_StopRec.setText(_translate("MainWindow", "Stop"))
        self.pushButton_Folder.setText(_translate("MainWindow", "Folder"))
        self.doubleSpinBox_TimeLapse.setSuffix(_translate("MainWindow", " sec"))
        self.pushButton_StartRec.setText(_translate("MainWindow", "Rec"))
        self.label.setText(_translate("MainWindow", "every"))
        self.label_7.setText(_translate("MainWindow", "Frame type:"))
        self.comboBox_FrameType.setItemText(0, _translate("MainWindow", "Light"))
        self.comboBox_FrameType.setItemText(1, _translate("MainWindow", "Dark"))
        self.comboBox_FrameType.setItemText(2, _translate("MainWindow", "Flat"))
        self.comboBox_FrameType.setItemText(3, _translate("MainWindow", "Bias"))
        self.action_DarkNoiseMeasurement.setText(_translate("MainWindow", "Dark Noise Measurement"))
from pyqtgraph import ImageView


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())



"""
CameraControl class
"""
import os.path
import shutil
import pickle
import numpy as np
import io
import re
import threading
import time
import datetime
from PyQt5 import QtCore


from picamera2 import Picamera2
from libcamera import controls, Rectangle
from . import pifcap_image

settings = {'name': 'camera hardware', 'type': 'group', 'children': [
    {
        'name': 'do hardware specific adjustments', 'type': 'bool', 'value': True,
        'tip': 'do adjustments for specific camera hardware',
    },
    {
        'name': 'force pixel size', 'type': 'bool', 'value': False, 'children': [
        {
            'name': 'X size', 'type': 'float', 'value': 1.0, 'suffix': ' µm', 'limits': [0.0, 10000.0],
        },
        {
            'name': 'Y size', 'type': 'float', 'value': 1.0, 'suffix': ' µm', 'limits': [0.0, 10000.0],
        },
    ]
    },
    {
        'name': 'force camera restarts', 'type': 'list', 'values': ['auto', 'yes', 'no'], 'value': 'auto',
        'tip': 'force camera restart after each exposure',
    },
]}


class CameraSettings:
    """exposure settings
    """

    def __init__(self):
        self._RawModes = list()
        self._selected_RawMode_idx = None
        self._is_newMode = True
        self._camera_controls = {
            "ExposureTime": 0.01,
            "AnalogueGain": 1.0,
            # disable all automatic regulations
            "AeEnable": False,
            "NoiseReductionMode": controls.draft.NoiseReductionModeEnum.Off,
            "AwbEnable": False,
            "ColourGains": (2.0, 2.0),  # to compensate the 2 G pixel in Bayer pattern
        }
        self._is_newControls = True
        # make thread safe
        self._AccessLock = QtCore.QMutex()


    @property
    def available_RawModes(self):
        self._AccessLock.lock()
        ret = self._RawModes
        self._AccessLock.unlock()
        return ret

    @available_RawModes.setter
    def available_RawModes(self, rms):
        self._AccessLock.lock()
        self._RawModes = rms
        self._AccessLock.unlock()
        self.set_RawModeFromIdx(0)

    @property
    def RawMode(self):
        self._AccessLock.lock()
        rm = self._RawModes[self._selected_RawMode_idx]
        self._AccessLock.unlock()
        return rm

    def set_RawModeFromIdx(self, idx):
        assert idx >= 0
        assert idx < len(self._RawModes)
        self._AccessLock.lock()
        if idx != self._selected_RawMode_idx:
            self._selected_RawMode_idx = idx
            self._is_newMode = True
            self._camera_controls["ExposureTime"] = min(
                self._camera_controls["ExposureTime"],
                self._RawModes[self._selected_RawMode_idx]["max_ExposureTime"]
            )
            self._camera_controls["ExposureTime"] = max(
                self._camera_controls["ExposureTime"],
                self._RawModes[self._selected_RawMode_idx]["min_ExposureTime"]
            )
            self._camera_controls["AnalogueGain"] = min(
                self._camera_controls["AnalogueGain"],
                self._RawModes[self._selected_RawMode_idx]["max_AnalogueGain"]
            )
            self._camera_controls["AnalogueGain"] = max(
                self._camera_controls["AnalogueGain"],
                self._RawModes[self._selected_RawMode_idx]["min_AnalogueGain"]
            )
            self._is_newControls = True
        self._AccessLock.unlock()

    @property
    def MaxExposureTime(self):
        self._AccessLock.lock()
        t = self._RawModes[self._selected_RawMode_idx]["max_ExposureTime"]
        self._AccessLock.unlock()
        return t / 1e6

    @property
    def MinExposureTime(self):
        self._AccessLock.lock()
        t = self._RawModes[self._selected_RawMode_idx]["min_ExposureTime"]
        self._AccessLock.unlock()
        return t / 1e6

    @property
    def MaxGain(self):
        self._AccessLock.lock()
        g = self._RawModes[self._selected_RawMode_idx]["max_AnalogueGain"]
        self._AccessLock.unlock()
        return g

    @property
    def MinGain(self):
        self._AccessLock.lock()
        g = self._RawModes[self._selected_RawMode_idx]["min_AnalogueGain"]
        self._AccessLock.unlock()
        return g

    @property
    def Binning(self):
        self._AccessLock.lock()
        b = self._RawModes[self._selected_RawMode_idx]["binning"]
        self._AccessLock.unlock()
        return b

    @property
    def ExposureTime(self):
        self._AccessLock.lock()
        t = self._camera_controls["ExposureTime"] / 1e6
        self._AccessLock.unlock()
        return t

    @ExposureTime.setter
    def ExposureTime(self, t):
        t_usec = int(round(1e6 * t))
        self._AccessLock.lock()
        t_usec = min(
            t_usec,
            self._RawModes[self._selected_RawMode_idx]["max_ExposureTime"]
        )
        t_usec = max(
            t_usec,
            self._RawModes[self._selected_RawMode_idx]["min_ExposureTime"]
        )
        if t_usec != self._camera_controls["ExposureTime"]:
            self._camera_controls["ExposureTime"] = t_usec
            self._is_newControls = True
        self._AccessLock.unlock()

    @property
    def Gain(self):
        self._AccessLock.lock()
        g = self._camera_controls["AnalogueGain"]
        self._AccessLock.unlock()
        return g

    @Gain.setter
    def Gain(self, g):
        self._AccessLock.lock()
        g = min(
            g,
            self._RawModes[self._selected_RawMode_idx]["max_AnalogueGain"]
        )
        g = max(
            g,
            self._RawModes[self._selected_RawMode_idx]["min_AnalogueGain"]
        )
        if g != self._camera_controls["AnalogueGain"]:
            self._camera_controls["AnalogueGain"] = g
            self._is_newControls = True
        self._AccessLock.unlock()

    @property
    def is_newControls(self):
        self._AccessLock.lock()
        ret = self._is_newControls
        self._AccessLock.unlock()
        return ret

    def reset_newControls(self):
        self._AccessLock.lock()
        self._is_newControls = False
        self._AccessLock.unlock()

    @property
    def is_newMode(self):
        self._AccessLock.lock()
        ret = self._is_newMode
        self._AccessLock.unlock()
        return ret

    def reset_newMode(self):
        self._AccessLock.lock()
        self._is_newMode = False
        self._AccessLock.unlock()

    @property
    def camera_controls(self):
        self._AccessLock.lock()
        ret = self._camera_controls
        self._AccessLock.unlock()
        return ret

    def __str__(self):
        return f'CameraSettings: RawMode={self.RawMode}, CameraControls={self.camera_controls}'

    def __repr__(self):
        return str(self)


class CameraControl(QtCore.QThread):
    """camera control and exposure thread
    """

    sigImage = QtCore.pyqtSignal(dict)
    sigRecordingFinished = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        # reset states
        self.picam2 = None
        self.CameraSettings = CameraSettings()
        self.CamProps = dict()
        self.needs_Restarts = False
        # exposure loop control
        self.Sig_ActionExit = threading.Event()  # abort and exit exposure loop
        self.Sig_CaptureDone = threading.Event()
        # exposure loop in separate thread
        self.Sig_ActionExit.clear()
        self.Sig_CaptureDone.clear()
        # handshake with GUI
        self.Sig_GiveImage = threading.Event()
        self.Sig_GiveImage.clear()
        # image recorder
        self._Folder = parent.Settings.get('recording', 'default folder')
        self._Comment = ""
        self._Prefix = ""
        self._FrameType = "Light"
        self._ImagesToRecord = 0
        self._ImagesRecorded = 0
        self._Record = False
        self._SettingsLock = QtCore.QMutex()
        self._Ext = ".pfc"

    def get_Cameras(self):
        """return list of available cameras"""
        cameras = Picamera2.global_camera_info()
        # use Id as unique camera identifier
        return [c["Model"] for c in cameras]

    def openCamera(self, idx: int):
        """open camera with given index idx
        """
        self.parent.log_Info("Opening camera.")
        self.picam2 = Picamera2(idx)
        # read camera properties
        self.CamProps = self.picam2.camera_properties
        self.parent.log_Info(f'Camera properties: {self.CamProps}')
        # force properties with values from config file
        if self.parent.Settings.get('camera hardware', 'force pixel size') or ("UnitCellSize" not in self.CamProps):
            self.parent.log_Info("Forcing UnitCellSize from program settings.")
            self.CamProps["UnitCellSize"] = (
                self.parent.Settings.get('camera hardware', 'force pixel size', 'X size'),
                self.parent.Settings.get('camera hardware', 'force pixel size', 'Y size'),
            )
        # some libcamera versions return a libcamera.Rectangle here!
        if type(self.CamProps["PixelArrayActiveAreas"][0]) is Rectangle:
            Rect = self.CamProps["PixelArrayActiveAreas"][0]
            self.CamProps["PixelArrayActiveAreas"] = (Rect.x, Rect.y, Rect.width, Rect.height)
        # raw modes
        self.CameraSettings.available_RawModes = self.getRawCameraModes()
        # some cameras need a restart after each exposure
        force_Restart = self.parent.Settings.get('camera hardware', 'force camera restarts')
        if force_Restart == "yes":
            self.parent.log_Info("Application setting forces camera restarts.")
            self.needs_Restarts = True
        elif force_Restart == "no":
            self.parent.log_Info("Application setting enabled camera restarts as needed.")
            self.needs_Restarts = False
        else:
            # "auto":
            self.needs_Restarts = self.CamProps["Model"] in ["imx290", "imx519"]
        # start exposure loop
        self.Sig_ActionExit.clear()
        self.start()

    def is_Open(self):
        """return camera state
        """
        return (self.picam2 is not None) and self.isRunning()

    def closeCamera(self):
        """close camera
        """
        self.parent.log_Info('Closing camera.')
        # stop exposure loop
        if self.isRunning():
            self.Sig_ActionExit.set()
            self.wait()  # wait until exposure loop exits
        # close picam2
        if self.picam2 is not None:
            if self.picam2.started:
                self.picam2.stop_()
            self.picam2.close()
        self.picam2 = None
        # reset states
        self.CamProps = dict()
        self.CameraSettings = CameraSettings()


    def reconfigure_Camera(self, RawMode, DoFastExposure=True):
        self.parent.log_Info(f'reconfiguring camera')
        config = self.picam2.create_still_configuration(
            queue=DoFastExposure,
            buffer_count=2 if DoFastExposure else 1  # need at least 2 buffer for queueing
        )
        # we do not need the main stream and configure it to smaller size to save memory
        config["main"]["size"] = (240, 190)
        # configure raw stream
        config["raw"] = {
            "size": RawMode["size"],
            "format": RawMode["camera_format"]
        }
        # optimize (align) configuration: small changes to some main stream configurations
        # (for instance: size) will fit better to hardware
        self.picam2.align_configuration(config)
        # set still configuration
        self.picam2.configure(config)

    def getRawCameraModes(self):
        """get list of usable raw camera modes
        """
        sensor_modes = self.picam2.sensor_modes
        raw_modes = []
        for sensor_mode in sensor_modes:
            # sensor_mode is dict
            # it must have key "format" (usually a packed data format) and can have
            # "unpacked" (unpacked data format)
            if "unpacked" not in sensor_mode.keys():
                sensor_format = sensor_mode["format"]
            else:
                sensor_format = sensor_mode["unpacked"]
            # packed data formats are not supported
            if sensor_format.endswith("_CSI2P"):
                self.parent.log_Warn(f'Raw mode not supported: {sensor_mode}!')
                continue
            # only monochrome and Bayer pattern formats are supported
            is_monochrome = re.match("R[0-9]+", sensor_format)
            is_bayer = re.match("S[RGB]{4}[0-9]+", sensor_format)
            if not (is_monochrome or is_bayer):
                self.parent.log_Warn(f'Raw mode not supported: {sensor_mode}!')
                continue
            #
            size = sensor_mode["size"]
            # adjustments for cameras:
            #   * zero- or garbage-filled columns
            #   * raw modes with binning or subsampling
            true_size = size
            binning = (1, 1)
            if self.parent.Settings.get('camera hardware', 'do hardware specific adjustments'):
                if self.CamProps["Model"] == 'imx477':
                    if size == (1332, 990):
                        true_size = (1332, 990)
                        binning = (2, 2)
                    elif size == (2028, 1080):
                        true_size = (2024, 1080)
                        binning = (2, 2)
                    elif size == (2028, 1520):
                        true_size = (2024, 1520)
                        binning = (2, 2)
                    elif size == (4056, 3040):
                        true_size = (4056, 3040)
                    else:
                        self.parent.log_Warn(f'Unsupported frame size {size} for imx477!')
                elif self.CamProps["Model"] == 'ov5647':
                    if size == (640, 480):
                        binning = (4, 4)
                    elif size == (1296, 972):
                        binning = (2, 2)
                    elif size == (1920, 1080):
                        pass
                    elif size == (2592, 1944):
                        pass
                    else:
                        self.parent.log_Warn(f'Unsupported frame size {size} for ov5647!')
                elif self.CamProps["Model"].startswith("imx708"):
                    if size == (1536, 864):
                        binning = (2, 2)
                    elif size == (2304, 1296):
                        binning = (2, 2)
                    elif size == (4608, 2592):
                        pass
                    else:
                        self.parent.log_Warn(f'Unsupported frame size {size} for imx708!')
            # add to list of raw formats
            raw_mode = {
                "label": f'{size[0]}x{size[1]} {sensor_format[1:5] if is_bayer else "mono"} {sensor_mode["bit_depth"]}bit',
                "size": size,
                "true_size": true_size,
                "camera_format": sensor_format,
                "bit_depth": sensor_mode["bit_depth"],
                "binning": binning,
            }
            self.reconfigure_Camera(raw_mode)
            ExposureTimeLimits = self.picam2.camera_controls['ExposureTime']
            AnalogueGainLimits = self.picam2.camera_controls['AnalogueGain']
            self.parent.log_Debug(f'Raw mode "{raw_mode["label"]}": ExposureTimeLimits={ExposureTimeLimits}, AnalogueGainLimits={AnalogueGainLimits}')
            raw_mode["min_ExposureTime"] = 0.0 if ExposureTimeLimits[0] is None else ExposureTimeLimits[0]
            raw_mode["max_ExposureTime"] = \
                1e6 if (ExposureTimeLimits[1] is None) or (ExposureTimeLimits[1] < ExposureTimeLimits[0]) \
                    else ExposureTimeLimits[1]
            raw_mode["min_AnalogueGain"] = 0.0 if AnalogueGainLimits[0] is None else AnalogueGainLimits[0]
            raw_mode["max_AnalogueGain"] = \
                250 if (AnalogueGainLimits[1] is None) or (AnalogueGainLimits[1] < AnalogueGainLimits[0]) \
                    else AnalogueGainLimits[1]
            raw_modes.append(raw_mode)
        # sort list of raw formats by size and bit depth in descending order
        raw_modes.sort(key=lambda k: k["size"][0] * k["size"][1] * 100 + k["bit_depth"], reverse=True)
        return raw_modes

    def getProp(self, name):
        """return camera properties
        """
        return self.CamProps[name]

    def log_FrameInformation(self, array, metadata, format):
        """write frame information to log

        Args:
            array: raw frame data
            metadata: frame metadata
            format: format string
        """
        if self.parent.config.getboolean("driver", "log_FrameInformation", fallback=False):
            if array.ndim == 2:
                arr = array.view(np.uint16)
                BitUsages = list()
                for b in range(15, -1, -1):
                    BitSlice = (arr & (1 << b)) != 0
                    BitUsage = BitSlice.sum() / arr.size
                    BitUsages.append(BitUsage)
                BitUsages = [f'{bu:.1e}' for bu in BitUsages]
                self.parent.log_Info(f'Frame format: {format}, shape: {array.shape} {array.dtype}, bit usages: (MSB) {" ".join(BitUsages)} (LSB)')
            else:
                self.parent.log_Info(f'Frame format: {format}, shape: {array.shape} {array.dtype}')
            self.parent.log_Info(f'Frame metadata: {metadata}')

    def set_RecordingSettings(self, Folder, Prefix, Comment, ImagesToRecord, FrameType):
        self._SettingsLock.lock()
        self._Folder = Folder
        self._Prefix = Prefix
        self._Comment = Comment
        self._ImagesToRecord = ImagesToRecord
        self._FrameType = FrameType
        self._SettingsLock.unlock()

    def Recording(self, Record):
        self._SettingsLock.lock()
        self._Record = Record
        if Record:
            self._ImagesRecorded = 0
        self._SettingsLock.unlock()

    def run(self):
        """exposure loop

        Made to run in a separate thread.
        """
        DoFastExposure = True
        while True:
            if self.Sig_ActionExit.is_set():
                # exit exposure loop
                self.picam2.stop_()
                return
            # picam2 needs to be open!
            if self.picam2 is None:
                raise RuntimeError("trying to make an exposure without camera opened")
            # need a camera stop/start when something has changed on camera configuration
            if self.CameraSettings.is_newMode or self.needs_Restarts:
                if self.picam2.started:
                    #self.parent.log_Info(f'Stopping camera for deeper reconfiguration.')
                    self.picam2.stop_()
                # change of DoFastExposure needs a configuration change
                self.CameraSettings.reset_newMode()
                self.reconfigure_Camera(RawMode=self.CameraSettings.RawMode, DoFastExposure=DoFastExposure)
            # changing exposure time or analogue gain can be done with camera running
            if self.CameraSettings.is_newControls:
                # change camera controls
                self.CameraSettings.reset_newControls()
                self.picam2.set_controls(self.CameraSettings.camera_controls)
            # start camera if not already running
            if not self.picam2.started:
                self.picam2.start()
                #self.parent.log_Info(f'Camera started.')
            # last chance to exit or abort before doing exposure
            if self.Sig_ActionExit.is_set():
                # exit exposure loop
                self.picam2.stop_()
                return
            # get (non-blocking!) frame and meta data
            self.Sig_CaptureDone.clear()  # set by on_CaptureFinished callback
            ExpectedEndOfExposure = time.time() + self.CameraSettings.ExposureTime
            job = self.picam2.capture_arrays(["raw"], wait=False, signal_function=self.on_CaptureFinished)
            PollingPeriod_s = 0.2
            while ExpectedEndOfExposure - time.time() > PollingPeriod_s:
                # allow to close camera
                if self.Sig_ActionExit.is_set():
                    # exit exposure loop
                    self.picam2.stop_()
                    return
                # allow exposure to finish earlier than expected (for instance when in fast exposure mode)
                if self.Sig_CaptureDone.is_set():
                    break
                time.sleep(PollingPeriod_s)
            # get frame and its metadata
            (array, ), metadata = self.picam2.wait(job)
            DateEnd = datetime.datetime.now(datetime.timezone.utc)
            #self.parent.log_Info('got exposed frame')
            # last chance to exit or abort before sending frame
            if self.Sig_ActionExit.is_set():
                # exit exposure loop
                self.picam2.stop_()
                return
            if not DoFastExposure:
                # in normal exposure mode the camera needs to be started with exposure command
                self.picam2.stop()
            # save image
            metadata = {
                k: metadata.get(k, None) 
                for k in [
                    'SensorBlackLevels', 'Lux', 'FrameDuration', 'DigitalGain',
                    'AnalogueGain', 'ScalerCrop', 'ExposureTime'
                ]
            }
            metadata["DateEnd"] = DateEnd
            metadata["FrameType"] = self._FrameType
            metadata["CameraModel"] = self.CamProps["Model"]
            metadata["UnitCellSize"] = self.CamProps["UnitCellSize"]
            metadata["Binning"] = self.CameraSettings.Binning
            self.on_Image(
                array=array, metadata=metadata, 
                format=self.picam2.camera_configuration()["raw"]["format"]
            )


    def on_CaptureFinished(self, Job):
        """callback function for capture done
        """
        self.Sig_CaptureDone.set()

    def on_Image(self, array, metadata, format):
        self._SettingsLock.lock()
        Img = pifcap_image.Image(array=array, metadata=metadata, format=format, comment=self._Comment)
        Record = self._Record
        ImagesRemain = self._ImagesToRecord - self._ImagesRecorded
        Folder = self._Folder
        self._SettingsLock.unlock()
        if Record and (ImagesRemain > 0):
            TimeStamp = datetime.datetime.now().strftime("%y%m%dT%H%M%S%f")[:16]
            self._SettingsLock.lock()
            FileName = os.path.join(Folder, f'{self._Prefix}-{TimeStamp}{self._Ext}')
            self._ImagesRecorded += 1
            if self._ImagesRecorded >= self._ImagesToRecord:
                self.sigRecordingFinished.emit()
                self._Record = False
            self._SettingsLock.unlock()
            Img.save(filename=FileName)
        disc_total, disc_used, disc_free = shutil.disk_usage(Folder)
        disc_free_images = disc_free // Img.estimate_FileSize()
        # send preview image to GUI
        if self.Sig_GiveImage.is_set():
            self.sigImage.emit({
                "array": array,
                "format": self.picam2.camera_configuration()["raw"]["format"],
                "metadata": metadata,
                "RecordingInfos": {
                    "ImagesRecorded": self._ImagesRecorded,
                    "disc_free": disc_free,
                    "disc_free_images": disc_free_images,
                },
            })
            self.Sig_GiveImage.clear()



"""
managing application settings, including a QtDialog to change them
"""

from PyQt5 import QtCore, QtGui, QtWidgets
import sys
import pyqtgraph.parametertree

from . import settings_ui


def name2key(name):
    """convert string to a valid QSettings key"""
    n = name
    for replacement in [(" ", "_")]:  # prepared for more replacements
        n = n.replace(*replacement)
    return n


class SettingsDialog(QtWidgets.QDialog):
    def __init__(self, settings, *args, **kwargs):
        super(SettingsDialog, self).__init__(*args, **kwargs)
        self.ui = settings_ui.Ui_Dialog()
        self.ui.setupUi(self)
        self.Settings = settings
        self.Params = pyqtgraph.parametertree.Parameter.create(name='params', type='group', children=settings)
        self.ui.ParamTree.setParameters(self.Params, showTop=False)

    def get(self, *names):
        return self.Params.child(*names).value()

    def store_QSettings(self):
        """stores all settings in QSettings"""
        # recursive function to traverse the tree
        def _traverseTree(QSettings, Params):
            for p in Params:
                if p.type() == "group":
                    QSettings.beginGroup(name2key(p.name()))
                    _traverseTree(QSettings, p)
                    QSettings.endGroup()
                else:
                    QSettings.setValue(name2key(p.name()), p.value())
        #
        QSettings = QtCore.QSettings()
        _traverseTree(QSettings, self.Params)

    def load_QSettings(self):
        """loads all settings from QSettings"""
        # recursive function to traverse the tree
        def _traverseTree(QSettings, Params):
            for p in Params:
                if p.type() == "group":
                    QSettings.beginGroup(name2key(p.name()))
                    _traverseTree(QSettings, p)
                    QSettings.endGroup()
                else:
                    p.setValue(QSettings.value(name2key(p.name()), p.defaultValue()))
        #
        QSettings = QtCore.QSettings()
        _traverseTree(QSettings, self.Params)



if __name__ == '__main__':
    ExampleSettings = [
        {'name': 'group 1', 'type': 'group', 'children': [
            {'name': 'an integer', 'type': 'int', 'value': 500, 'limits': [100, 10000],
             'tip': 'this is an integer'},
            {'name': 'a string', 'type': 'str', 'value': "test string", 'tip': 'this is a string'},
            {'name': 'a float', 'type': 'float', 'value': 3.0, 'limits': [2.0, 3600], 'step': 0.5, 'suffix': ' sec',
             'tip': 'this is a float'},
            {'name': 'a 2nd float', 'type': 'float', 'value': 6.0, 'tip': 'this is a 2nd float'},
            {'name': 'a bool', 'type': 'bool', 'value': True, 'tip': "this is a checkbox"},
        ]},
        {'name': 'group 2', 'type': 'group', 'children': [
            {'name': 'a bool with nested float', 'type': 'bool', 'value': False, 'children': [
                {'name': u'nested float', 'type': 'float', 'value': 23.0, 'limits': [-400.0, 80.0],
                 'step': 1.0, 'visible': True},
            ]},
        ]},
        {'name': 'a hidden float', 'type': 'float', 'value': 10.0, 'limits': [2.0, 3600], 'step': 0.5, 'suffix': ' m',
         'tip': 'this is a hidden float', 'visible': False},
    ]
    #
    App = QtWidgets.QApplication(sys.argv)
    App.setOrganizationName("GeierSoft")
    App.setOrganizationDomain("Astro")
    App.setApplicationName("settings")
    #
    #
    MainWindow = SettingsDialog(settings=ExampleSettings)
    # MainWindow.resize(1400, 900)
    MainWindow.show()
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        sys.exit(App.exec_())



"""
Fast Raspberry Pi image capturing software for Lucky Imaging.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
import sys
import os
import os.path
import numpy as np
import pyqtgraph as pg

from . import pifcap_ui
from . import camera
from . import settings
from . import autoexposure

__author__ = "Ronald Schreiber"
__copyright__ = "Copyright 2024"
__version__ = "v1.0.0"
__license__ = "MIT"
__title__ = f'pifcap {__version__}'

theme_selection = "Dark"  # "Dark", "Light"

# Interpret image data as row-major instead of col-major
pg.setConfigOptions(imageAxisOrder='row-major')


class MainWin(QtWidgets.QMainWindow):
    """
    pifcap main window.
    """

    def __init__(self, showDebugMessages=False):
        super(MainWin, self).__init__()
        # settings
        DefaultSettings = [
            {'name': 'saturation limit', 'type': 'int', 'value': 95, 'limits': [0, 100], 'suffix': ' %',
             'tip': 'mark saturated pixel when above % of full-scale',
            },
            autoexposure.settings,
            camera.settings,
            {'name': 'recording', 'type': 'group', 'children': [
                {
                    'name': 'default folder', 'type': 'str', 'value': os.path.join(os.path.expanduser("~"), "Pictures"),
                    'tip': 'default folder to store images',
                },
            ]},
        ]
        self.Settings = settings.SettingsDialog(parent=self, settings=DefaultSettings)
        self.Settings.load_QSettings()
        # GUI
        pg.setConfigOption('background', pg.mkColor(100, 0, 0))
        self.showDebugMessages = showDebugMessages
        self.ui = pifcap_ui.Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle(__title__)
        self.ui.plainTextEdit_Log.setMaximumBlockCount(100)
        self.ui.pushButton_Settings.clicked.connect(self.on_Settings_clicked)
        self.ui.groupBox_CameraSettings.setEnabled(False)
        self.ui.pushButton_StartRec.setEnabled(False)
        self.ui.comboBox_Camera.setEnabled(True)
        #
        self.ui.ImageView_Preview.setLevels(0, 2**16 - 1)
        self.ui.ImageView_Preview.ui.roiBtn.hide()
        self.ui.ImageView_Preview.ui.menuBtn.hide()
        # saturated pixel highlight
        self.SaturatedPixelImage = pg.ImageItem()
        self.SaturatedPixelImage.setZValue(65000)
        self.ui.ImageView_Preview.addItem(self.SaturatedPixelImage)
        #
        self.ImageFolder = self.Settings.get('recording', 'default folder')
        # camera
        self.Cam = camera.CameraControl(parent=self)
        self.Cam.sigImage.connect(self.on_Image)
        self.Cam.sigRecordingFinished.connect(self.on_pushButton_StopRec_clicked)
        self.Cam.Sig_GiveImage.set()
        self.ui.comboBox_Camera.addItems(self.Cam.get_Cameras())
        # recording settings
        self.ui.lineEdit_ImageComment.textChanged.connect(self.on_RecordingSettingsChanged)
        self.ui.lineEdit_FileNamePrefix.textChanged.connect(self.on_RecordingSettingsChanged)
        self.ui.spinBox_nImagesToRecord.valueChanged.connect(self.on_RecordingSettingsChanged)
        self.ui.doubleSpinBox_TimeLapse.valueChanged.connect(self.on_RecordingSettingsChanged)
        self.ui.comboBox_FrameType.currentIndexChanged.connect(self.on_RecordingSettingsChanged)
        # preview image
        self.Img = None
        # exposure time optimization
        self.AutoExposure = autoexposure.AutoExposure(Settings=self.Settings)

    def add_LogMessage(self, Msg, Severity="INFO"):
        if Severity=="ERROR":
            Html = '<font color="red">ERROR: %s</font>' % Msg
        elif Severity=="WARN":
            Html = '<font color="orange">WARN: %s</font>' % Msg
        elif (Severity=="DEBUG"):
            if self.showDebugMessages:
                Html = '<font color="yellow">DBG: %s</font>' % Msg
            else:
                return
        else: # INFO
            #Html = '<font color="black">%s</font>' % Msg
            Html = Msg
        self.ui.plainTextEdit_Log.appendHtml(Html)
        self.ui.statusbar.showMessage(Msg)
        self.ui.plainTextEdit_Log.ensureCursorVisible()

    def log_Error(self, Msg):
        self.add_LogMessage(Msg, "ERROR")

    def log_Warn(self, Msg):
        self.add_LogMessage(Msg, "WARN")

    def log_Debug(self, Msg):
        self.add_LogMessage(Msg, "DEBUG")

    def log_Info(self, Msg):
        self.add_LogMessage(Msg, "INFO")

    @QtCore.pyqtSlot()
    def closeEvent(self, event):
        self.add_LogMessage("entering closeEvent", Severity="DEBUG")
        self.Cam.closeCamera()
        #self.Settings.save_QSettings()  # FIXME: activate this!
        # proceed with close
        event.accept()

    def on_Settings_clicked(self):
        self.Settings.show()

    @QtCore.pyqtSlot()
    def on_pushButton_ConnectDisconnect_clicked(self):
        if self.Cam.is_Open():
            self.Cam.closeCamera()
            self.ui.pushButton_ConnectDisconnect.setText("connect")
            #self.ui.comboBox_RawMode.clear()
            self.ui.groupBox_CameraSettings.setEnabled(False)
            self.ui.pushButton_StartRec.setEnabled(False)
            self.ui.comboBox_Camera.setEnabled(True)
        else:
            self.Cam.openCamera(self.ui.comboBox_Camera.currentIndex())
            self.ui.comboBox_Camera.setEnabled(False)
            self.ui.comboBox_RawMode.clear()
            self.ui.comboBox_RawMode.addItems([rm["label"] for rm in self.Cam.CameraSettings.available_RawModes])
            self.ui.comboBox_RawMode.setCurrentIndex(0)
            self.ui.pushButton_ConnectDisconnect.setText("disconnect")
            self.ui.groupBox_CameraSettings.setEnabled(True)
            self.ui.pushButton_StartRec.setEnabled(True)


    @QtCore.pyqtSlot(int)
    def on_comboBox_RawMode_currentIndexChanged(self, idx):
        if idx >= 0:
            self.Cam.CameraSettings.set_RawModeFromIdx(idx)
            self.ui.doubleSpinBox_Gain.setMaximum(self.Cam.CameraSettings.MaxGain)
            self.ui.doubleSpinBox_Gain.setMinimum(self.Cam.CameraSettings.MinGain)
            self.ui.doubleSpinBox_Gain.setValue(self.Cam.CameraSettings.Gain)
            self.ui.doubleSpinBox_ExposureTime.setMaximum(self.Cam.CameraSettings.MaxExposureTime)
            self.ui.doubleSpinBox_ExposureTime.setMinimum(self.Cam.CameraSettings.MinExposureTime)
            self.ui.doubleSpinBox_ExposureTime.setValue(self.Cam.CameraSettings.ExposureTime)


    @QtCore.pyqtSlot(float)
    def on_doubleSpinBox_ExposureTime_valueChanged(self, t):
        self.Cam.CameraSettings.ExposureTime = t

    @QtCore.pyqtSlot(float)
    def on_doubleSpinBox_Gain_valueChanged(self, g):
        self.Cam.CameraSettings.Gain = g

    @QtCore.pyqtSlot()
    def on_pushButton_Folder_clicked(self):
        ImageFolder = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Folder to store images..",
            self.ImageFolder
        )
        if ImageFolder != '':
            self.ImageFolder = ImageFolder
        self.on_RecordingSettingsChanged()

    def on_RecordingSettingsChanged(self):
        self.Cam.set_RecordingSettings(
            Folder=self.ImageFolder,
            Prefix=self.ui.lineEdit_FileNamePrefix.text(),
            Comment=self.ui.lineEdit_ImageComment.text(),
            ImagesToRecord=self.ui.spinBox_nImagesToRecord.value(),
            FrameType=self.ui.comboBox_FrameType.currentText(),
        )

    @QtCore.pyqtSlot(int)
    def on_checkBox_OptimizeExposure_stateChanged(self, state):
        if state:
            self.AutoExposure.init_optimize()

    @QtCore.pyqtSlot(dict)
    def on_Image(self, Img):
        self.ui.label_RecordingInfos.setText(
            f'{Img["RecordingInfos"]["disc_free"]/1024/1024:.0f} MiB (~{Img["RecordingInfos"]["disc_free_images"]} images) free'
        )
        self.ui.label_RecordedImageCounter.setText(f'{Img["RecordingInfos"]["ImagesRecorded"]} images saved')
        self.log_Debug(f'received image: {Img["format"]}, {Img["array"].shape}, {Img["array"].dtype}')
        self.log_Debug(f'received metadata: {Img["metadata"]}')
        self.Img = Img
        self.process_Image()
        # TODO: wie wird Aufnahme automatisch gestoppt?
        QtCore.QTimer.singleShot(100, self.request_NewImage)

    def request_NewImage(self):
        self.Cam.Sig_GiveImage.set()

    def process_Image(self):
        format = self.Img["format"]
        array = self.Img["array"]
        # we expect uncompressed format here
        if format.count("_") > 0:
            raise NotImplementedError(f'got unsupported raw image format {format}')
        if format[0] not in ["S", "R"]:
            raise NotImplementedError(f'got unsupported raw image format {format}')
        # Bayer or mono format
        if format[0] == "S":
            # Bayer pattern format
            BayerPattern = format[1:5]
            # FIXME: BayerPattern = self.parent.config.get("driver", "force_BayerOrder", fallback=BayerPattern)
            bit_depth = int(format[5:])
        else:
            # mono format
            BayerPattern = None
            bit_depth = int(format[1:])
        # left adjust if needed
        if bit_depth > 8:
            array = array.view(np.uint16)
        else:
            array = array.view(np.uint8)
        #
        # remove 0- or garbage-filled columns # TODO implement this?
        #true_size = self.present_CameraSettings.RawMode["true_size"]
        #array = array[0:true_size[1], 0:true_size[0]]
        # exposure time optimization
        if self.ui.checkBox_OptimizeExposure.isChecked():
            finished, newExposureTime = self.AutoExposure.do_optimize(
                array=array,
                ExposureTime=self.Img["metadata"]["ExposureTime"]/1e6,
                n_bits=bit_depth,
                MinExposureTime=self.ui.doubleSpinBox_ExposureTime.minimum(),
                MaxExposureTime=self.ui.doubleSpinBox_ExposureTime.maximum()
            )
            if finished:
                self.ui.checkBox_OptimizeExposure.setChecked(False)
            else:
                self.ui.doubleSpinBox_ExposureTime.setValue(newExposureTime)
        # color channels
        imgR = None
        imgG1 = None
        imgG2 = None
        imgB = None
        if BayerPattern is None:
            # mono format
            img = array
        else:
            BayerChannels = {
                "GBRG": {"G1": (0, 0), "B": (0, 1), "R": (1, 0), "G2": (1, 1)},
                "BGGR": {"B": (0, 0), "G1": (0, 1), "G2": (1, 0), "R": (1, 1)},
            }[BayerPattern]
            imgR = array[BayerChannels["R"][0]::2, BayerChannels["R"][1]::2]
            imgG1 = array[BayerChannels["G1"][0]::2, BayerChannels["G1"][1]::2]
            imgG2 = array[BayerChannels["G2"][0]::2, BayerChannels["G2"][1]::2]
            imgB = array[BayerChannels["B"][0]::2, BayerChannels["B"][1]::2]
            RawPreviewMode = self.ui.comboBox_RawPreviewMode.currentText()
            if RawPreviewMode == "luminance":
                img = (0.2126 * imgR + 0.7152 / 2 * imgG1 + 0.7152 / 2 * imgG2 + 0.0722 * imgB).astype(int)
            elif RawPreviewMode == "red":
                img = imgR
            elif RawPreviewMode == "green 1":
                img = imgG1
            elif RawPreviewMode == "green 2":
                img = imgG2
            elif RawPreviewMode == "blue":
                img = imgB
            else:
                raise NotImplementedError
        # preview rotation and flip
        Rot = self.ui.comboBox_PreviewRot.currentText()
        FlipH = self.ui.checkBox_PreviewFlipH.isChecked()
        FlipV = self.ui.checkBox_PreviewFlipV.isChecked()
        img = self.flipRotate_Image(img, Rot, FlipH, FlipV)
        # show preview image
        self.ui.ImageView_Preview.setImage(img, autoLevels=False,
                                            autoHistogramRange=False,
                                            #levelMode="rgb",
                                            autoRange=False,
                                            )
        # highlight saturated pixel
        if self.ui.checkBox_Saturation.isChecked():
            sat = np.zeros((img.shape[0], img.shape[1], 4), dtype=int)
            sat[:, :, 0] = 1  # pure red, full transparent
            satLim = (2**bit_depth) * self.Settings.get('saturation limit') / 100
            if BayerPattern is None:
                # mono image
                is_sat = img > satLim
            else:
                is_sat = (imgR > satLim) | (imgG1 > satLim) | (imgG2 > satLim) | (imgB > satLim)
            is_sat = self.flipRotate_Image(is_sat, Rot, FlipH, FlipV)
            sat[:, :, 3] = is_sat * 1  # saturated pixel get intransparent red
            self.SaturatedPixelImage.setImage(sat, levels=[[0, 1], [0, 1], [0, 1], [0, 1]])
        else:
            self.SaturatedPixelImage.clear()

    def flipRotate_Image(self, img, Rot, FlipH, FlipV):
        if Rot != "0deg":
            k = {"90deg": 3, "180deg": 2, "270deg": 1}[Rot]
            img = np.rot90(img, k=k)
        if FlipH:
            img = np.fliplr(img)
        if FlipV:
            img = np.flipud(img)
        return img

    @QtCore.pyqtSlot()
    def on_pushButton_StartRec_clicked(self):
        self.ui.pushButton_StopRec.setEnabled(True)
        self.ui.pushButton_StartRec.setEnabled(False)
        self.Cam.Recording(True)

    @QtCore.pyqtSlot()
    def on_pushButton_StopRec_clicked(self):
        self.Cam.Recording(False)
        self.ui.pushButton_StartRec.setEnabled(True)
        self.ui.pushButton_StopRec.setEnabled(False)




# Start Qt event loop unless running in interactive mode.
def main():
    import argparse
    parser = argparse.ArgumentParser(description=f'{__title__}, a fast Raspberry Pi image capturing software for Lucky Imaging.')
    parser.add_argument("-d", "--debug", action="store_true",
                        help="enable debug messages")
    args = parser.parse_args()
    # build application
    App = QtWidgets.QApplication(sys.argv)
    App.setOrganizationName("GeierSoft")
    App.setOrganizationDomain("Astro")
    App.setApplicationName("pifcap")
    #
    # copied from https://stackoverflow.com/questions/48256772/dark-theme-for-qt-widgets
    if theme_selection == 'Dark':
        App.setStyle("Fusion")
        #
        # # Now use a palette to switch to dark colors:
        dark_palette = QtGui.QPalette()
        dark_palette.setColor(QtGui.QPalette.Window, QtGui.QColor(53, 53, 53))
        dark_palette.setColor(QtGui.QPalette.WindowText, QtCore.Qt.white)
        dark_palette.setColor(QtGui.QPalette.Base, QtGui.QColor(35, 35, 35))
        dark_palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))
        dark_palette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(25, 25, 25))
        dark_palette.setColor(QtGui.QPalette.ToolTipText, QtCore.Qt.white)
        dark_palette.setColor(QtGui.QPalette.Text, QtCore.Qt.white)
        dark_palette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
        dark_palette.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.white)
        dark_palette.setColor(QtGui.QPalette.BrightText, QtCore.Qt.red)
        dark_palette.setColor(QtGui.QPalette.Link, QtGui.QColor(42, 130, 218))
        dark_palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(42, 130, 218))
        dark_palette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(35, 35, 35))
        dark_palette.setColor(QtGui.QPalette.Active, QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
        dark_palette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, QtCore.Qt.darkGray)
        dark_palette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, QtCore.Qt.darkGray)
        dark_palette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.Text, QtCore.Qt.darkGray)
        dark_palette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.Light, QtGui.QColor(53, 53, 53))
        App.setPalette(dark_palette)
    elif theme_selection == 'Light':
        App.setStyle("")
        pass
    else:
        pass
    #
    MainWindow = MainWin(showDebugMessages=args.debug)
    # MainWindow.resize(1400, 900)
    MainWindow.show()
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        sys.exit(App.exec_())


if __name__ == '__main__':
    main()



"""
command line program to convert pifcap internal image format to FITS
"""

import sys
import os.path
import pickle
import glob
import numpy as np
import argparse
import multiprocessing
import time
import datetime
from astropy.io import fits


def convert(input_filename, remove, skip_existing):
    output_filename = input_filename.rsplit(".", maxsplit=1)[0] + ".fits"
    if not(skip_existing and os.path.isfile(output_filename)):
        # convert
        retrials = 20
        while True:
            try:
                with open(input_filename, "rb") as fh:
                    img = pickle.load(fh)
            except EOFError:
                retrials -= 1
                if retrials > 0:
                    time.sleep(0.1)
                else:
                    print(f'ERROR: Can not open {input_filename}', file=sys.stderr)
                    return input_filename
            else:
                break
        array = img["array"]
        format = img["format"]
        metadata = img["metadata"]
        # we expect uncompressed format here
        if format.count("_") > 0:
            raise NotImplementedError(f'got unsupported raw image format {format}')
        if format[0] not in ["S", "R"]:
            raise NotImplementedError(f'got unsupported raw image format {format}')
        # Bayer or mono format
        if format[0] == "S":
            # Bayer pattern format
            BayerPattern = format[1:5]
            bit_depth = int(format[5:])
        else:
            # mono format
            BayerPattern = None
            bit_depth = int(format[1:])
        # left adjust if needed
        if bit_depth > 8:
            bit_pix = 16
            array = array.view(np.uint16)
        else:
            bit_pix = 8
            array = array.view(np.uint8)
        # convert to FITS
        hdu = fits.PrimaryHDU(array)
        hdu.header["BZERO"] = (2 ** (bit_pix - 1), "offset data range")
        hdu.header["BSCALE"] = (1, "default scaling factor")
        hdu.header["ROWORDER"] = ("TOP-DOWN", "Row order")
        hdu.header["INSTRUME"] = (metadata["CameraModel"], "CCD Name")
        hdu.header["EXPTIME"] = (metadata["ExposureTime"]/1e6, "[s] Total Exposure Time")
        hdu.header["DATE-OBS"] = (
            (metadata["DateEnd"] - datetime.timedelta(seconds=metadata["ExposureTime"]/1e6)).isoformat(timespec="milliseconds"),
            "UTC time of observation start"
        )
        hdu.header["DATE-END"] = (metadata["DateEnd"].isoformat(timespec="milliseconds"), "UTC time at end of observation")
        #hdu.header["CCD-TEMP"] = (metadata.get('SensorTemperature', 0), "[degC] CCD Temperature")
        hdu.header["PIXSIZE1"] = (metadata["UnitCellSize"][0] / 1e3, "[um] Pixel Size 1")
        hdu.header["PIXSIZE2"] = (metadata["UnitCellSize"][1] / 1e3, "[um] Pixel Size 2")
        hdu.header["XBINNING"] = (metadata["Binning"][0], "Binning factor in width")
        hdu.header["YBINNING"] = (metadata["Binning"][1], "Binning factor in height")
        hdu.header["XPIXSZ"] = (metadata["UnitCellSize"][0] / 1e3 * metadata["Binning"][0], "[um] X binned pixel size")
        hdu.header["YPIXSZ"] = (metadata["UnitCellSize"][1] / 1e3 * metadata["Binning"][1], "[um] Y binned pixel size")
        hdu.header["FRAME"] = (metadata["FrameType"], "Frame Type")
        hdu.header["IMAGETYP"] = (metadata["FrameType"] + " Frame", "Frame Type")
        hdu.header["GAIN"] = (metadata["AnalogueGain"], "Gain")
        if BayerPattern is not None:
            hdu.header["XBAYROFF"] = (0, "[px] X offset of Bayer array")
            hdu.header["YBAYROFF"] = (0, "[px] Y offset of Bayer array")
            hdu.header["BAYERPAT"] = (BayerPattern, "Bayer color pattern")
        if "SensorBlackLevels" in metadata:
            SensorBlackLevels = metadata["SensorBlackLevels"]
            if len(SensorBlackLevels) == 4:
                # according to picamera2 documentation:
                #   "The black levels of the raw sensor image. This
                #    control appears only in captured image
                #    metadata and is read-only. One value is
                #    reported for each of the four Bayer channels,
                #    scaled up as if the full pixel range were 16 bits
                #    (so 4096 represents a black level of 16 in 10-
                #    bit raw data)."
                # When image data is stored as 16bit it is not needed to scale SensorBlackLevels again.
                # But when we store image with 8bit/pixel we need to divide by 2**8.
                SensorBlackLevelScaling = 2 ** (bit_pix - 16)
                hdu.header["OFFSET_0"] = (SensorBlackLevels[0] * SensorBlackLevelScaling, "[DN] Sensor Black Level 0")
                hdu.header["OFFSET_1"] = (SensorBlackLevels[1] * SensorBlackLevelScaling, "[DN] Sensor Black Level 1")
                hdu.header["OFFSET_2"] = (SensorBlackLevels[2] * SensorBlackLevelScaling, "[DN] Sensor Black Level 2")
                hdu.header["OFFSET_3"] = (SensorBlackLevels[3] * SensorBlackLevelScaling, "[DN] Sensor Black Level 3")
        hdu.header["comment"] = img["comment"]
        hdul = fits.HDUList([hdu])
        # save FITS
        hdul.writeto(output_filename, overwrite=True)
        # remove input if requested
        if remove:
            os.remove(input_filename)
    return input_filename


def main():
    # command line parser
    parser = argparse.ArgumentParser(description='convert pifcap to FITS')
    parser.add_argument('-s', '--skip_existing', action="store_true",
                        help='skip conversion when output exists')
    parser.add_argument('-r', '--remove', action="store_true",
                        help='remove input file after conversion')
    parser.add_argument('-d', '--demon', action="store_true",
                        help='convert existing files, stay running and convert all upcoming new files; exit with CTRL-C')
    parser.add_argument('-j', '--jobs', type=int, default=1,
                        help='number of parallel running conversion jobs (default: 1)')
    parser.add_argument('-v', '--verbose', action="store_true",
                        help='verbose messages')
    parser.add_argument('files', metavar="path", default='*.pfc',
                        help='files to convert; accepts "?", "*" and character ranges like "[a-z]" (default "*.pfc")')
    args = parser.parse_args()
    if args.demon:
        print("Demon mode. Exit with CTRL+C.")
    # file lists
    pending_files = list()
    pending_processes = list()
    finished_files = list()
    # multiprocessing pool
    pool = multiprocessing.Pool(processes=args.jobs)
    # demon loop
    while True:
        # start conversion for all new input files
        all_filenames = glob.glob(args.files)
        for fn in all_filenames:
            if (fn not in finished_files) and (fn not in pending_files):
                pending_processes.append(
                    pool.apply_async(
                        convert,
                        kwds={
                            "input_filename": fn,
                            "remove": args.remove,
                            "skip_existing": args.skip_existing,
                        }
                    )
                )
                pending_files.append(fn)
        # tell what is going on
        if args.verbose:
            print(f'{len(pending_processes)} files pending, {len(finished_files)} files converted   ', end='\r')
        # release CPU
        time.sleep(0.1)
        # pending process loop
        while len(pending_processes) > 0:
            still_pending_processes = list()
            for pp in pending_processes:
                if pp.ready():
                    finished_file = pp.get()
                    finished_files.append(finished_file)
                    pending_files.remove(finished_file)
                else:
                    still_pending_processes.append(pp)
            pending_processes = still_pending_processes
            if args.demon:
                # demon mode allows to start more processes before pending are done
                break
            else:
                # non-demon: stay in pending process loop
                if args.verbose:
                    print(f'{len(pending_processes)} files pending, {len(finished_files)} files converted   ', end='\r')
                time.sleep(0.1)
        # non-demon: finished
        if not args.demon:
            break
    print()


if __name__ == '__main__':
    main()
    sys.exit(0)
